/* house telephone - connects two POTS devices
 *
 * (w) 2017 Axel Schwenke <axel.schwenke@gmx.de>
 *
 * $Id: main.c 379 2017-12-22 13:31:35Z schwenke $
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/wdt.h>

/* Pin Usage:
                    ATTiny13A
                  ------------
 n.c. .... Reset | 1        8 | Vcc ..... +5V
 Sense .... ADC3 | 2        7 | SCK ..... n.c.
 n.c. ..... ADC2 | 3        6 | MISO .... Talk
 GND ....... GND | 4        5 | MOSI .... Ring
                  ------------
*/

// sense voltage levels
#include <math.h>   //for floor()

#define UREF  1.1f  //reference voltage (nominal 1.1V)
#define RH    470   //voltage divider upper resistor
#define RL    10    //voltage divider lower resistor
#define V1    33    //below 33V -> ring
#define V2    20    //below 20V -> talk
#define V3    37    //above 37V -> idle

// those are evaluated at compile time
#define L1    ((uint16_t)floor(V1/UREF*RL/(RH+RL)*1024))
#define L2    ((uint16_t)floor(V2/UREF*RL/(RH+RL)*1024))
#define L3    ((uint16_t)floor(V3/UREF*RL/(RH+RL)*1024))

// ADC samples multiple times
#define OVERSAMPLE 4


// the state machine has 4 states
typedef enum { IDLE, WAIT, RING, TALK } state_t;

// time keeping, ticks down @ 50Hz
volatile uint8_t tick;


// Timer0 interrupt every 20ms = 50Hz
ISR(TIM0_COMPA_vect)
{
    uint8_t i;
    if ((i = tick)) {
        tick= i-1;
    }
}


// hardware initialization
static void hw_init(void)
{
    //clock prescaler = 16 -> 600kHz
    CLKPR = (1<<CLKPCE);
    CLKPR = (1<<CLKPS2);

    //turn on ADC, clock divider=4 (150kHz)
    ADCSRA = (1<<ADEN | 1<<ADPS1);
    //internal reference, ADC3
    ADMUX  = (1<<REFS0 | 1<<MUX1 | 1<<MUX0);
    //disconnect ADC3 from digital buffer
    DIDR0  = (1<<ADC3D);
    //start a dummy conversion
    ADCSRA |= (1<<ADSC);

    //Timer0: CTC mode
    //overflow every 20ms = 12.000 clock ticks
    TCCR0A = (1<<WGM01);
    //prescaler = 64
    TCCR0B = (1<<CS01 | 1<<CS00);
    //divide by 188 -> 20.05ms
    OCR0A  = 187;
    //enable overflow IRQ
    TIMSK0 = (1<<OCIE0A);

    //enable PB0 + PB1 output pins & set to L
    //unused pins are inputs with pullup enabled
    DDRB   =  (1<<PB1 | 1<<PB0);
    PORTB  = ~(1<<PB3 | 1<<PB1 | 1<<PB0);

    //enable idle sleeping
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_enable();

    //enable the watchdog
    wdt_enable(WDTO_1S);

    //wait for ADC
    while (ADCSRA & (1<<ADSC)) ;
}


// switch hardware to "RING" mode
static inline void ring_on(void)
{
    PORTB  |= (1<<PB1);
    TCCR0A |= (1<<COM0A0);
}


// switch hardware to "TALK" mode
static inline void talk_on(void)
{
    TCCR0A &= ~(1<<COM0A0);
    PORTB   = PORTB & ~(1<<PB1 | 1<<PB0);
}


// measure voltage at "sense" pin
static uint16_t AD_conversion(void) {
    uint16_t result= 0;
#if (OVERSAMPLE > 1)
    for (uint8_t i=0; i<OVERSAMPLE; i++) {
#endif
        ADCSRA |= (1<<ADSC);
        while (ADCSRA & (1<<ADSC)) ;
        result += ADC;
#if (OVERSAMPLE > 1)
    }
    result /= OVERSAMPLE;
#endif

    return result;
}


// and here we go
int __attribute__((OS_main)) main(void)
{
    hw_init();
    sei();

    //wait 50 ticks = 1s
    tick= 50;
    while (tick) {
        wdt_reset();
        sleep_cpu();
    }

    state_t state= IDLE;

    while (1) {
        uint16_t sense= AD_conversion();

        switch (state) {
        case IDLE:
            if (sense < L1) {
                state= WAIT;
            }
            break;

        case WAIT:
            if (sense < L2) {
                state= TALK;
            } else if (sense > L3 ) {
                state= IDLE;
            } else if (tick == 0) {
                state= RING;
                ring_on();
                tick= 20;
            }
            break;

        case RING:
            if (tick == 0) {
                state= WAIT;
                talk_on();
                tick= 80;
            }
            break;

        case TALK:
            if (sense > L3) {
                state= IDLE;
            }
            break;
        }

        //sleep until next timer overflow
        wdt_reset();
        sleep_cpu();
    }
}

